/* Any frequently used shortcuts, strings and constants */

/* encourage jslint not to worry about global soundManager */
/*global soundManager: false */
var sm = soundManager;

/* PreviewPlayer class constructor */
function PreviewPlayer() {
    PreviewPlayer.superclass.constructor.apply(this, arguments);
}

PreviewPlayer.NAME = "previewPlayer";

/** hash of static CSS class names
 *
 * @property CSS_CLASSES
 * @type {Object}
 * @private
 * @static
 */
PreviewPlayer.CSS_CLASSES = {
    previewLink: ".preview",
    previewCtrl: ".preview .ctrl",
    previewText: ".preview .acchide",
    previewPlayClass: "preview-play",
    previewPauseClass: "preview-pause",
    rowSel: "li",
    trackNum: ".track-num",
    trackNumContainer: ".num",
    vu: ".vu",
    vuLeft: ".left",
    vuRight: ".right",
    progressBar: ".progressbar",
    progressPlay: ".play",
    progressLoad: ".load"
};

PreviewPlayer.ATTRS = {
    /**
     * width of row as apposed to width of contentBox
     * TODO consider using default 'width' ATTR
     */
    rowWidth: {
        value: "" // default value of player width
    },
    oneError: {
        value: "one:error"
    }
};

PreviewPlayer.HTML_PARSER = {
    rowWidth: function(srcNode) {
        return srcNode.get('offsetWidth');
    }
};

Y.extend(PreviewPlayer, Y.Widget, {

    CONTENT_TEMPLATE: null,

    initializer: function() {
        // Use a cache object to stash our song objects for re-use
        // really just a way to enforce a limit on how many items
        // we keep.
        this.cache = new Y.Cache({max:10, uniqueKeys: true});
    },

    renderUI : function() {
        var vuTemplate = Y.one("#tpl-vu-meter"),
            progressbarTemplate = Y.one("#tpl-progressbar");

        this.set("vuMeter", Y.Node.create(vuTemplate.getContent()));
        this.set("progress", Y.Node.create(progressbarTemplate.getContent()));
    },

    bindUI : function() {
        var srcNode = this.get("srcNode");
        srcNode.delegate("click", this.handleClick, PreviewPlayer.CSS_CLASSES.previewLink, this);


        Y.on("windowresize", Y.bind(this.handleResize, this));

        this.after('rowWidthChange', this.afterTrWidthChange);

        // perform Unity integration if required.
        Y.Unity.onInit(Y.bind(this.bindUnity, this));
    },

    // Beyond this point is the PreviewPlayer specific application and rendering logic

    handleClick: function(e) {
        e.halt();
        var target = e.currentTarget,
            currentSound = this.get("currentSound"),
            sound = this.getOrCreateSound(target);

        if (sound === currentSound) {
            sound.togglePause();
        } else {
            if (currentSound) {
                currentSound.stop();
            }
            if (sound) {
                sound.play();
            }
        }
    },

    /**
     * update the rowWidth attribute when a resize event is fired
     */
    handleResize: function() {
        var srcNode = this.get("srcNode");
        return this.set('rowWidth', srcNode.one('li').get('offsetWidth'));
    },
    /**
     * handle change to rowWidth attr
     */
    afterTrWidthChange: function() {
        var bar = this.get("progressBar");
        if (bar) {
            bar.setStyle("width", this.get("rowWidth")+"px");
        }
    },
    getTrackId: function(node) {
        return "u1ms" + node.getAttribute("data-u1ms-trackid");
    },
    getTrackUrl: function(node){
        return node.getAttribute("href");
    },
    getOrCreateSound: function(node) {
        var id = this.getTrackId(node),
            url = this.getTrackUrl(node),
            sound, cachedSound;

        if (id && url) {
            cachedSound = this.cache.retrieve(id);
            if (cachedSound === null) {
                sound = sm.createSound({
                    id: id,
                    url: url,
                    onpause: this.soundPause,
                    onplay: this.soundPlay,
                    onresume: this.soundPlay,
                    onstop: this.soundStop,
                    whileloading: this.soundWhileLoading,
                    whileplaying: this.soundWhilePlaying,
                    onload: this.soundOnload,
                    onfinish: this.soundFinish,
                    multiShot: false,
                    usePeakData: true
                });
                sound.u1msData = {
                    node: node,
                    widget: this,
                    vu: {
                        node: null,
                        leftVu: null,
                        rightVu: null
                    },
                    throbbing: false
                };
                this.cache.add(id, sound);
            } else {
                sound = cachedSound.response;
            }
            return sound;
        }
    },
    /*
    * Preload the next track in the list
    */
    getFirstNode: function() {
        var srcNode = this.get("srcNode");
        return srcNode.one(PreviewPlayer.CSS_CLASSES.previewLink);
    },

    getNextNode: function(currentNode) {
        var nextRow,
            currentRow;

        currentRow = currentNode.ancestor(PreviewPlayer.CSS_CLASSES.rowSel);
        nextRow = currentRow.next(PreviewPlayer.CSS_CLASSES.rowSel);

        if (nextRow) {
            return nextRow.one(PreviewPlayer.CSS_CLASSES.previewLink);
        }

        return false;
    },
    getPrevNode: function(currentNode) {
        var prevRow,
            currentRow;

        currentRow = currentNode.ancestor(PreviewPlayer.CSS_CLASSES.rowSel);
        prevRow = currentRow.previous(PreviewPlayer.CSS_CLASSES.rowSel);

        if (prevRow) {
            return prevRow.one(PreviewPlayer.CSS_CLASSES.previewLink);
        }

        return false;
    },
    setActiveRow: function(currentNode, sound) {
        var currentRow = currentNode.ancestor(PreviewPlayer.CSS_CLASSES.rowSel),
            previewText, previewCtrl, vu, vuLeft, vuRight, progress;

        currentRow.addClass("active");
        previewText = currentRow.one(PreviewPlayer.CSS_CLASSES.previewText);
        if (previewText) {
            previewText.setContent("Pause");
        }
        previewCtrl = currentRow.one(PreviewPlayer.CSS_CLASSES.previewCtrl);
        if (previewCtrl) {
            previewCtrl.removeClass(PreviewPlayer.CSS_CLASSES.previewPlayClass);
            previewCtrl.addClass(PreviewPlayer.CSS_CLASSES.previewPauseClass);
        }


        if (!this.get("progressBar")) {
            progress = this.get("progress");

            this.set("progressBar", progress.one(PreviewPlayer.CSS_CLASSES.progressBar));
            this.set("playMeter", progress.one(PreviewPlayer.CSS_CLASSES.progressPlay));
            this.set("loadMeter", progress.one(PreviewPlayer.CSS_CLASSES.progressLoad));

            currentRow.appendChild(progress);
        }

        if (sound.isHTML5) {
            currentRow.appendChild('<span class="vu vu-dummy"></span>');
        } else {
            vu = this.get("vuMeter");
            if (vu) {
                vuLeft = vu.one(PreviewPlayer.CSS_CLASSES.vuLeft);
                vuRight = vu.one(PreviewPlayer.CSS_CLASSES.vuRight);
                if (vuLeft  && vuRight) {
                    vuLeft.setStyle("height", 0);
                    vuRight.setStyle("height", 0);
                }
                currentRow.appendChild(vu);
            }
        }
    },
    clearActiveRow: function() {
        var srcNode = this.get("srcNode"),
            currentRow = srcNode.one(".active"),
            currentNode, previewText, previewCtrl, vu, progress, pm, lm;
        if (currentRow) {
            
            currentNode = currentRow.one(PreviewPlayer.CSS_CLASSES.previewLink);

            currentRow.removeClass("active");

            // set meter to zero early so it doesn't appear
            // in the next track as the old value
            pm = this.get("playMeter");
            if (pm) {
                pm.setStyle('width', '0');
            }

            lm = this.get("loadMeter");
            if (lm) {
                lm.setStyle('width', '0');
            }

            previewText = currentRow.one(PreviewPlayer.CSS_CLASSES.previewText);
            if (previewText) {
                previewText.setContent("Play");
            }
            previewCtrl = currentRow.one(PreviewPlayer.CSS_CLASSES.previewCtrl);
            if (previewCtrl) {
                previewCtrl.removeClass(PreviewPlayer.CSS_CLASSES.previewPauseClass);
                previewCtrl.addClass(PreviewPlayer.CSS_CLASSES.previewPlayClass);
            }
            vu = currentRow.one(PreviewPlayer.CSS_CLASSES.vu);
            if (vu) {
                vu.remove();
            }
            progress = currentRow.one(PreviewPlayer.CSS_CLASSES.progressBar);
            if (progress) {
                progress.remove();
            }
        }
    },
    showThrobber: function (sound) {
        var currentRow = sound.u1msData.node.ancestor(PreviewPlayer.CSS_CLASSES.rowSel);
        currentRow.addClass('loading');
        sound.u1msData.throbbing = true;
    },
    removeThrobber: function (sound) {
        var currentRow,
            throbber;
        
        // if we're not throbbing, do not do anything
        if(sound.u1msData.throbbing === false) {
            return;
        }

        currentRow = sound.u1msData.node.ancestor(PreviewPlayer.CSS_CLASSES.rowSel);
        currentRow.removeClass('loading');
        
        // if theres a throbber, remove it
        throbber = currentRow.one('.throbber');
        if(throbber) {
            throbber.remove();
        }
        
        sound.u1msData.throbbing = false;
    },
    soundPlay: function() {
        // context is that of the sound object
        var that = this.u1msData.widget,
            currentNode = this.u1msData.node,
            mp,
            trackInfo;

        that.set("currentSound", this);
        that.clearActiveRow();
        
        // song not loaded, show throbber
        if(this.readyState !== 3) {
            that.showThrobber(this);
        }
        
        that.setActiveRow(currentNode, this);
        if (Y.Unity.ready) {
            mp = Y.Unity.Unity.MediaPlayer;
            mp.setPlaybackState(mp.PlaybackState.PLAYING);
            trackInfo = {
                title: currentNode.getAttribute("data-u1ms-title"),
                artist: currentNode.getAttribute("data-u1ms-artist"),
                album: currentNode.getAttribute("data-u1ms-release"),
                artLocation: currentNode.getAttribute("data-u1ms-coverart")
            };
            mp.setTrack(trackInfo);
            // And show a notification
            Y.Unity.Unity.Notification.showNotification(
                trackInfo.title,
                "by " + trackInfo.artist + " from " + trackInfo.album,
                trackInfo.artLocation);
        }
    },
    soundResume: function() {
        // context is that of the sound object
        var that = this.u1msData.widget,
            currentNode = this.u1msData.node;
        that.clearActiveRow();
        that.setActiveRow(currentNode, this);
    },
    soundPause: function() {
        // context is that of the sound object
        var that = this.u1msData.widget,
            mp;

        that.clearActiveRow();
        if (Y.Unity.ready) {
            mp = Y.Unity.Unity.MediaPlayer;
            mp.setPlaybackState(mp.PlaybackState.PAUSED);
        }
    },
    soundStop: function() {
        // context is that of the sound object
        var that = this.u1msData.widget,
            mp;

        that.clearActiveRow();
        if (Y.Unity.ready) {
            mp = Y.Unity.Unity.MediaPlayer;
            mp.setPlaybackState(mp.PlaybackState.PAUSED);
        }
    },
    soundFinish: function() {
        // context is that of the sound object
        var that = this.u1msData.widget,
            currentNode = this.u1msData.node,
            next = that.getNextNode(currentNode),
            nextSound;
        if (next) {
            nextSound = that.getOrCreateSound(next);
            nextSound.play();
        } else {
            that.clearActiveRow();
        }
    },
    soundWhileLoading: function() {
        var that = this.u1msData.widget;
        that.updateLoadMeter(this.bytesLoaded, this.bytesTotal);
    },
    soundWhilePlaying: function() {
        // context is that of the sound object
        var that = this.u1msData.widget;
        
        // in html5 mode this is called prematurely
        // so skip that fired event altogheter
        if(this.isHTML5 && this.position === 0) {
            return;
        }
        
        // throbber should never exist while playing
        that.removeThrobber(this);
        
        if (sm.flashVersion >= 9 && this.instanceOptions.usePeakData) {
            that.updatePeakData(this);
        }
        that.updatePlayMeter(this.position, this.durationEstimate);
    },
    updatePeakData: function(sound) {
        var vu = sound.u1msData.vu;

        if (vu.node) {
            vu.vuLeft.setStyle('height', Math.ceil(17*sound.peakData.left)+'px');
            vu.vuRight.setStyle('height', Math.ceil(17*sound.peakData.right)+'px');
        } else {
            // vus not set
            vu.node = sound.u1msData.node.ancestor('li').one(PreviewPlayer.CSS_CLASSES.vu);
            vu.vuLeft = vu.node.one(PreviewPlayer.CSS_CLASSES.vuLeft);
            vu.vuRight = vu.node.one(PreviewPlayer.CSS_CLASSES.vuRight);
        }
        
    },
    /**
     * update the display of the load meter
     * set width to 100% when we're done, so it adjusts to resize events
     * @param current {Number} bytes loaded
     * @param total {Number} total bytes to load
     */
    updateLoadMeter: function(current, total) {
        var width = this.get("rowWidth"),
            ratio;

        ratio = (current/total);

        if (ratio === 1) {
            width = '100%';
        } else {
            width = Math.floor(ratio * width) + 'px';
        }
        this.get("loadMeter").setStyle("width", width);
    },
    /**
     * update the display of the play meter
     * @param current {Number} bytes played
     * @param total {Number} total bytes to play
     */
    updatePlayMeter: function(current, total) {
        var width = this.get('rowWidth');

        width *= (current/total);
        this.get('playMeter').setStyle('width', Math.floor(width) +'px');
    },

    // Preload the next song
    soundOnload: function() {
        // context is that of the sound object
        var that = this.u1msData.widget,
            currentNode = this.u1msData.node,
            cachedSound = that.cache.retrieve(that.getTrackId(currentNode));
        
        if(cachedSound.response.readyState !== 2) {
            // sound loaded
            this.onposition(this.duration - 5000, Y.bind(function() {
                var nextNode = this.getNextNode(currentNode);
                if (nextNode) {
                    this.getOrCreateSound(nextNode);
                }
            }, that));
        }
        else {
            // sound not loaded (readyState 2 == failed/error)
            Y.Global.fire(that.get('oneError'), {message: "The selected song failed to load. Please try again later."});
            
            // destroy the bad SM object
            cachedSound.response.destruct();
            
            // make this cache item go away now
            cachedSound.expires = 1;
            
            // unset currentSound
            that.set('currentSound', undefined);
            
            // remove active classes and such
            that.clearActiveRow();
        }
    },

    bindUnity: function() {
        Y.log("Configuring Unity integration");
        var mp = Y.Unity.Unity.MediaPlayer;
        mp.init("Ubuntu One Music");
        mp.setCanPlay(true);
        mp.setCanPause(true);
        mp.setCanGoNext(true);
        mp.setCanGoPrevious(true);
        mp.setPlaybackState(mp.PlaybackState.PAUSED);
        mp.onPlayPause(Y.bind(this.unityPlayPause, this));
        mp.onNext(Y.bind(this.unityNext, this));
        mp.onPrevious(Y.bind(this.unityPrevious, this));
    },

    unityPlayPause: function() {
        Y.log("Sound Menu: play/pause");
        var currentSound = this.get("currentSound"),
            node;
        if (currentSound) {
            currentSound.togglePause();
            return;
        }
        // No current sound: start playing the first track.
        node = this.getFirstNode();
        if (node) {
            currentSound = this.getOrCreateSound(node);
            currentSound.play();
        }
    },

    unityNext: function() {
        Y.log("Sound Menu: next track");
        var currentSound = this.get("currentSound"), node;
        if (currentSound) {
            currentSound.stop();
            node = this.getNextNode(currentSound.u1msData.node);
        } else {
            node = this.getFirstNode();
        }
        if (node) {
            currentSound = this.getOrCreateSound(node);
            currentSound.play();
        }
    },

    unityPrevious: function() {
        Y.log("Sound Menu: previous track");
        var currentSound = this.get("currentSound"), node;
        if (currentSound) {
            currentSound.stop();
            node = this.getPrevNode(currentSound.u1msData.node);
        } else {
            node = this.getFirstNode();
        }
        if (node) {
            currentSound = this.getOrCreateSound(node);
            currentSound.play();
        }
    }

});

Y.namespace("U1MS").PreviewPlayer = PreviewPlayer;
