/**
* DownloadStatus class constructor
*
* TODO Refactor: don't touch the DOM so much, store values in local object
*
* Class to monitor a list of downloads represented in rows of a table,
* state is stored as data attributes on those rows. Once all downloads have
* reached a 'completed state' polling stops but can be restarted at user
* request by clicking on a table row.
*
* @constructor
*/
/* Any frequently used shortcuts, strings and constants */
var ONEERROR = "one:error";


function DownloadStatus() {
    DownloadStatus.superclass.constructor.apply(this, arguments);
}

DownloadStatus.NAME = "downloadStatus";

DownloadStatus.ATTRS = {
    dataNodes: { value: null },

    pollingInterval: {
        value: 10000, // 10 seconds
        validator: function(val) {
            return Y.Lang.isNumber(val);
        },
        setter: function(val) {
            this.pollingInterval = val;
            return val;
        },
        lazyAdd: false
    },

    // comepleted states used as a based for when to stop polling.
    completedStates: {
        value: [
            "complete",
            "error",
            "unknown",
            "downloaded-not-present"
        ]}
};

// Get nodes within the srcNode.
DownloadStatus.HTML_PARSER = {
    // DataNodes which will have the data attrs.
    dataNodes: ['tbody tr']
};

Y.extend(DownloadStatus, Y.Widget, {

    // set up instance specific state
    initializer: function(config) {
        if (config.statusCheckUrl && config.statusCheckUrl.length) {
            this.statusCheckUrl = config.statusCheckUrl;
        } else {
            throw new Error("one-musicstore-status: need a status check url");
        }
    },

    bindUI: function() {
        var bb = this.get("boundingBox");
        this.pollNotification = bb.one(".notification");
        this.pollNotification.one(".cta").on("click", function() {
            this.checkStatus(false);
        }, this);
    },

    syncUI: function() {
        this.interval = null;
        this.setPollTimer();
        this.checkStatus(false);
    },

    destructor: function() {
        clearInterval(this.interval);
    },

    /**
     * Enum for css class name values
     * @readonly
     * @enum {string}
     */
    SELECTORS: {
    // The data attribute for the dld key
        dldIdAttrKey: "data-dld-id",
        // The data attrubyte for the status.
        dldStatusAttrKey: "data-dld-status",
        // Description node selector
        dldStatusDescSel: ".dld-status-desc",
        // Where to find the class prefixed by classStatusPrefix
        dldStatusClassSel: ".dld-status"
    },

    /**
     * DownloadStatus's common class prefix for statuses.
     * @type {string}
     */
    CLASS_STATUS_PREFIX: "status-",

    /**
     * Maximum poll count
     * @type {number}
     */
    POLL_MAX: 30,

    /**
     * Poll count store for downloads
     * @type {Object}
     */
    pollCount: 0,

    setPollTimer: function() {
        this.interval = setInterval(Y.bind(this.checkStatus, this), this.pollingInterval);
    },

    /**
     * parse the nodes for status data
     */
    getStatusData: function() {

        var dldData = {},
        dldIdAttrKey = this.SELECTORS.dldIdAttrKey,
        completedStates = Y.Array(this.get("completedStates"));

        this.get("dataNodes").each(function(node) {
            var key = node.getAttribute(dldIdAttrKey),
            dldStatusAttrKey = this.SELECTORS.dldStatusAttrKey,
            nodeStatus = node.getAttribute(dldStatusAttrKey);

            // don't ask about stuff which is polled out or completed
            if ( Y.Array.indexOf(completedStates, nodeStatus) === -1) {
                dldData[key] = {};
                dldData[key].status = this.parseNode(node);
            }
        }, this);

        return dldData;
    },

    parseNode: function(node) {
        var dldStatusAttrKey = this.SELECTORS.dldStatusAttrKey;
        return node.getAttribute(dldStatusAttrKey);
    },

    checkStatus: function(poll) {
        // Get data for all the downloads
        if (this.pollCount < this.POLL_MAX || poll === false) {
            var data = this.getStatusData();

            if (Y.Object.isEmpty(data)) {
                this.removeTimer();
                return;
            }

            this.postDownloadData(data);
            this.pollCount++;
        } else {
            this.removeTimer();
            this.pollNotification.addClass("show");
        }
    },

    /**
     * postDownloadData. Send the download data...
     *
     * @param {Object} dldData all the data on the downloads
     * @param {?Object} io Y.IO object
     */
    postDownloadData: function(dldData, io) {
        io = io || new Y.IO();
        var cfg = {
            context: this,
            data: {'song_list': Y.JSON.stringify(dldData),
                'csrfmiddlewaretoken': YUI.Env.CSRF_TOKEN},
                method: "POST",
                on: {
                    "success": this.onSuccess,
                    "failure": this.onFailure
                }

        }, request;
        request = io.send(this.statusCheckUrl, cfg);
    },

    /**
     * onSuccess
     *
     * Take the response data from the statusCheck service and store the
     * values in the attributes of the download elements, as well as
     * setting relevant classes and content
     *
     */
    onSuccess: function(tid, response) {
        var respData = Y.JSON.parse(response.responseText),
            srcNode = this.get("srcNode"),
            dldIdAttrKey = this.SELECTORS.dldIdAttrKey,
            dldStatusAttrKey = this.SELECTORS.dldStatusAttrKey,
            continuePolling = false,
            completedStates = Y.Array(this.get("completedStates")),
            key,
            selector,
            dataNode,
            oldStatus,
            descNode,
            descStatusClassNode,
            nodeStatus,
            classStatusPrefix;

        for (key in respData) {
            selector = '['+ dldIdAttrKey + '="' + key + '"]';
            dataNode = srcNode.one(selector),
            oldStatus = dataNode.getAttribute(dldStatusAttrKey),
            descStatusClassNode = dataNode.one(this.SELECTORS.dldStatusClassSel),
            descNode = dataNode.one(this.SELECTORS.dldStatusDescSel),
            nodeStatus = respData[key].status.toLowerCase(),
            classStatusPrefix = this.CLASS_STATUS_PREFIX;


            if (dataNode) {
                if (oldStatus !== nodeStatus) {

                    if (descStatusClassNode) {
                        // Remove current status
                        descStatusClassNode.removeClass(classStatusPrefix + oldStatus);
                        // Replace with new status
                        descStatusClassNode.addClass(classStatusPrefix + nodeStatus);
                    }

                    // Update description
                    if (descNode) {
                        descNode.set("text", respData[key].status_desc);
                    }
                    // Update status attr
                    dataNode.setAttribute(dldStatusAttrKey, nodeStatus);
                }
            }


            if ( Y.Array.indexOf(completedStates, nodeStatus) === -1 ) {
                continuePolling = true;
            }
        }

        // if we're all complete, remove the poll timer
        if (!continuePolling) {
            this.removeTimer();
        }

        if (!continuePolling && this.pollNotification && this.pollNotification.hasClass("show")) {
            this.pollNotification.removeClass("show");
        }

        // Return continuePolling to add testing.
        return continuePolling;
    },

    onFailure: function () {
        Y.Global.fire(ONEERROR);
    },

    /**
     * remove the timer (stop polling) once we've reached POLL_MAX or
     * all items are at a complete status
     */
    removeTimer: function() {
        if (this.interval) {
            clearInterval(this.interval);
            delete this.interval;
        }
    }

});

Y.namespace("U1MS").DownloadStatus = DownloadStatus;
