/*
 * This module adds some basic support for writing out CSS styles into a style-block in the page
 * this can be useful for avoiding minimising reflows/repaints where you want to change the
 * styles of multiple elements that are generated in JS.
 *
 * Example:
 *
 * YUI().use('one-stylenode', function(Y) {
 *     var myStyles = new Y.u1.StyleNode([".first { color: red }", ".first { font-size: 200% }", ".second { color: blue }"]);
 *     window.setTimeout(function(){
 *         myStyles.find(".first").setStyle("color", "yellow");
 *     }, 2000);
* });
 *
 * Notes:
 *      find() returns a rule as a node object. You can then use setStyle to modify the values.
 *      findAll() returns all matching rules as a nodeList.
 *
 */

Y.namespace("u1");

var StyleNode = function() {
    StyleNode.superclass.constructor.apply(this, arguments);
};

Y.mix(StyleNode, {
    NAME: "stylenode",
    ATTRS: {
        "styleNode": {value: null },
        "styleSheet": {value: null }
    }
});

Y.extend(StyleNode, Y.Base, {
    initializer: function(aRules) {
        aRules = aRules || null;
        if (aRules) {
            this.create(aRules);
        }
    },
    create: function(aRules) {
        var styleFrag = Y.Node.create("<style type='text/css'></style>"),
            ss,
            i,
            j,
            curSheet,
            owner;

        Y.Node.one("head").append(styleFrag);
        ss = styleFrag.get("styleSheet");
        if (ss) {
            ss.cssText = aRules.join(" ");
        } else {
            styleFrag.setContent(aRules.join(" "));
        }
        styleFrag.set("id", Y.guid());
        for (i = 0, j = document.styleSheets.length; i < j; i++) {
            curSheet = document.styleSheets[i],
                owner = curSheet.ownerNode || curSheet.owningElement;
            if (owner.id === styleFrag.get("id")) {
                this.set("styleSheet", curSheet);
            }
        }
        this.set("styleNode", styleFrag);
    },
    find: function(selector) {
        var styleSheet = this.get("styleSheet"),
        i,
        rules;

        if (styleSheet) {
            rules = (styleSheet.cssRules || styleSheet.rules);
            for (i = 0; i < rules.length; i++) {
                if (rules[i].selectorText === selector) {
                    return Y.Node(rules[i]);
                }
            }
        }
        return null;
    },
    findAll: function(selector) {
        var styleSheet = this.get("styleSheet"),
            rules = (styleSheet.cssRules || styleSheet.rules),
            nList = new Y.NodeList(),
            i;

        for (i = 0; i < rules.length; i++) {
            if (rules[i].selectorText === selector) {
                nList.push(Y.Node(rules[i]));
            }
        }
        return nList;
    }
});
Y.u1.StyleNode = StyleNode;
