
var UploaderTimeRemaining;

/**
* reports speed of upload, time remaining, pr upload session
*
* @module one-uploader-measure
*/
UploaderTimeRemaining = Y.Base.create('oneUploaderTimeRemaining', Y.Base, [], {

    _tr_timer: null,

    _startStamp: null,

    initializer: function () {

        if(this.get('calculateTimeRemaining')) {
            this.after('render', this._tr_bind);
        }

        /**
        * Fires every second with the following payload
        *   statistics: the statistics attribute
        *   remaining: the readable time remaining
        *   rawRemaining: time remaining in ms
        *
        * @event fullycomplete
        */
        this.publish('timeremaining');

    },


    _tr_bind: function () {
        this.on('uploadstart', this._tr_start, this);
        this.on('alluploadscomplete', this._tr_stop);
        this.on('totaluploadprogress', this._tr_progress);

    },

    _tr_unbind: function () {
        this.detach('uploadstart', this._tr_start);
        this.detach('alluploadscomplete', this._tr_stop);
        this.detach('totaluploadprogress', this._tr_progress);
    },

    _tr_start: function () {
        this._startStamp = new Date();

        this._tr_timer = Y.later(1000, this, this._tr_poll, {}, true);
    },

    _tr_stop: function () {

        if(this._tr_timer) {
            this._tr_timer.cancel();
        }

        this._startStamp = null;

    },

    _tr_progress: function (e) {

        this.set('statistics.total', e.bytesTotal);
        this.set('statistics.loaded', e.bytesLoaded);

    },

    _tr_poll: function () {

        this.set('statistics.rate', this.getTransferRate());
        this.set('statistics.remaining', this.getTimeRemaining());


        this.fire('timeremaining', {
            statistics:     this.get('statistics'),
            remaining:      this.getReadableTimeRemaining(),
            rawRemaining:   this.getTimeRemaining()
        });

    },

    destructor: function () {
        this._tr_stop();
        this._tr_unbind();
    },

    /**
    * Returns time remaining in ms
    *
    * @method getTimeRemaining
    * @public
    */
    getTimeRemaining: function () {
        var total = this.get('statistics.total'),
            loaded = this.get('statistics.loaded'),
            rate = this.get('statistics.rate');

        if(total && loaded && rate) {
            // do not trust the YUI Uploader to report correct numbers
            // convert negative times to 0
            return Math.max(0, (total - loaded) / rate);
        } else {
            return null;
        }

    },

    /**
    * Returns time remaining in a human readable string (1 second/2 days)
    *
    * @method getReadableTimeRemaining
    * @public
    */
    getReadableTimeRemaining: function () {
        var s = this.get('statistics.remaining'),
            ret = {};

        if(!s) {
            return null;
        }

        // convert to seconds
        s = s / 1000;

        if(s < 60) {
            ret.unit = this.get('strings.time_s');
            ret.val = Math.floor(s);
        } else if(s >= 60 && s < 3600) {
            ret.unit = this.get('strings.time_m');
            ret.val = Math.floor((s / 60));
        } else if(s >= 3600 && s < 86400) {
            ret.unit = this.get('strings.time_h');
            ret.val = Math.floor((s / 3600));
        } else if(s >= 86400) {
            ret.unit = this.get('strings.time_d');
            ret.val = Math.floor((s / 86400));
        } else {
            // unknown
            ret.unit = this.get('strings.time_u');
            ret.val = null;
        }

        if(ret.val !== 1) {
            ret.unit += this.get('strings.time_plural');
        }
        return ret.val + ' ' + ret.unit;
    },

    /**
    * Calculate transfer rates using a EMA (exponential moving average)
    *
    * @method getTransferRate
    * @public
    */
    getTransferRate: function () {
        var loaded      = this.get('statistics.loaded'),
            elapsed     = this.getTimeElapsed(),
            lastRate,
            rate        = this.get('statistics.rate'),
            sF          = this.get('smoothingFactor');

        if(loaded > 0 && elapsed > 0) {

            lastRate = loaded / elapsed;

            // initialize rate with lastRate
            if(!rate) {
                this.set('statistics.rate', lastRate);
                rate = lastRate;
            }
            return sF * lastRate + (1 - sF) * rate;
        } else {
            return 0;
        }

    },

    /**
    * Calculates time passed since upload started
    *
    * @method getTimeElapsed
    * @public
    */
    getTimeElapsed: function () {
        var ret;
        if(this._startStamp) {
            ret = new Date() - this._startStamp;
        } else {
            ret = null;
        }
        return ret;
    }

}, {

    ATTRS: {
        /**
        * Use this module
        *
        * @attribute calculateTimeRemaining
        * @type {Boolean}
        * @default true
        */
        calculateTimeRemaining: { value: true },

        /**
        * Smoothing factor (how much weight the previous value carries)
        * Should be a number between 1 and 0
        *
        * @attribute smoothingFactor
        * @type {Number}
        * @default 0.005
        */
        smoothingFactor: {
            value: 0.005,
            validator: function (val) {
                // numbers 0 and 1 does not make sense
                return Y.Lang.isNumber(val) && val > 0 && val < 1;
            }
        },

        /**
        * Statistic data
        *
        * @attribute statistics
        * @type {Object}
        */
        statistics: {
            value: {
                total: null,
                loaded: null,
                remaining: null,
                rate: null
            }
        }

    }

});

Y.namespace('U1.Widgets').UploaderTimeRemaining = UploaderTimeRemaining;
