
var Uploader,

    UPLOADING = 'UPLOADING', // file is being uploaded
    STARTED = 'STARTED', // file has been started
    COMPLETE = 'COMPLETE', // file has been fully uploaded
    FAILED = 'FAILED', // file failed to upload
    CANCELLED = 'CANCELLED', // file cancelled

    BINDUI          = 'bindUI',
    SYNCUI          = 'syncUI';

/**
* Extends uploader-html5 with a frontend and functionality.
*
* @module one-uploader
*/
Uploader = Y.Base.create('oneUploader', Y.UploaderHTML5, [], {

    initializer: function () {

        Y.before(this._bindUploaderEvents, this, BINDUI);
        Y.before(this._syncUploader, this, SYNCUI);

        window.onbeforeunload = Y.bind(this._beforeUnload, this);

        /**
        * Fires when the UI has no more uploads left
        * This is different than the queue reporting it is done,
        * since the queue can skip files
        *
        * @event fullycomplete
        */
        this.publish('fullycomplete');

        /**
        * Fires when the UI has one or more uploads left
        * Exists because of queues reporting done even if uploads failed
        *
        * @event fullycomplete
        */
        this.publish('partialcomplete');

        /**
        * Fires when all uploads were cancelled, or failed
        *
        * @event incomplete
        */
        this.publish('incomplete');

        this.set('status', this.get('strings.uploads_empty'));
        this.set('dropAreaStatus', this.get('strings.drop_ready'));

    },

    destructor : function() {

        window.onbeforeunload = null;

    },

    _beforeUnload: function (e) {
        var msg = this._getUnloadMessage();
        if(msg) {
            e.returnValue = msg;
            return msg;
        }
    },

    /**
    * Binds all upload events, as well as some UI events
    *
    * @method _bindUploaderEvents
    */
    _bindUploaderEvents: function () {

        this._bindDD();

        this.after(['filecancel', 'fileselect'],    this._totalUploadProgress);

        this.on('totaluploadprogress',              this._totalUploadProgress);
        this.on('alluploadscomplete',               this._dispatchComplete);
        this.on('uploadstart',                      this._uploaderStart);

        this.on('uploadprogress',                   this._uploadProgress);
        this.on('uploadcomplete',                   this._uploadComplete);
        this.on('uploaderror',                      this._uploadError);
        this.on('filecancel',                       this._uploadCancel);
        this.on('fileuploadstart',                  this._uploadStart);

        this.after('dragAndDropAreaChange',         this._bindDD);

    },

    /**
    * Sets state of UI
    *
    * @method _syncUploader
    */
    _syncUploader: function () {

        this.fire('statusChange');
        this.fire('dropAreaStatusChange');

    },

    _startDrag: function () {

        this.set('dropAreaStatus', this.get('strings.drop_pending'));
        this.get('contentBox').addClass(this.getClassName('drop-pending'));

    },

    _endDrag: function () {

        this.set('dropAreaStatus', this.get('strings.drop_ready'));
        this.get('contentBox').removeClass(this.getClassName('drop-pending'));

    },

    _bindDD: function (event) {
        var ev = event || {prevVal: null},
            ddArea = this.get('dragAndDropArea');

        if (ev.prevVal !== null) {
            this.detach('drop');
            this.detach('dragenter');
            this.detach('dragover');
            this.detach('dragleave');
        }

        if (ddArea !== null) {
            this.on(['dragenter', 'dragover'], this._startDrag);
            this.on(['dragleave', 'drop'], this._endDrag);
        }

    },

    /**
    * Handler for the user starting uploads
    *
    * @oaram event {Event}
    * @method _uploaderStart
    * @protected
    */
    _uploaderStart: function () {

        this.set('enabled', false);

    },

    _uploadProgress: function (e) {

        this._setFileStatus(e.file, UPLOADING);

    },

    _uploadComplete: function (e) {

        this._setFileStatus(e.file, COMPLETE);

    },

    _uploadError: function (e) {

        this._setFileStatus(e.file, FAILED);

    },

    _uploadCancel: function (e) {

        this._setFileStatus(e.file, CANCELLED);

    },

    _uploadStart: function (e) {

        this._setFileStatus(e.file, STARTED);

    },

    _setFileStatus: function (file, status) {

        if(!file.get('status')) {
            file.addAttr('status', {
                value: status
            });
        }

        file.set('status', status);

    },

    _getUnloadMessage: function () {
        var msg;

        if(this.getHaveUploadsInProgress()) {

            msg = Y.Lang.sub(this.get('strings.cancel_remaining'), { count: this.queue.queuedFiles.length});

        } else if(this.getFilesNotUploaded().length > 0) {

            msg = Y.Lang.sub(this.get('strings.cancel_pending'), { count: this.getFilesNotUploaded().length });

        }

        return msg;
    },

    /**
    * Main path for destroying the uploader
    * Will ask the user nicely what to do
    * if there are uploads remaining/ongoing etc
    *
    * @method _confirmCancel
    * @protected
    */
    _confirmCancel: function () {

        var msg = this._getUnloadMessage();

        if(this.getHaveUploadsInProgress()) {

            if (window.confirm(msg)) {

                // make sure we nuke the xhrs
                this.cancelAllUploads();

                this.destroy();
            }

        } else if(this.getFilesNotUploaded().length > 0) {

            if (window.confirm(msg)) {
                this.destroy();
            }

        } else {
            this.destroy();
        }

    },

    _skipStalledUploads: function () {

        if(this.getHaveUploadsInProgress()) {

            this.queue.restartUpload();

        } else {

            this.set('status', Y.Lang.sub(this.get('strings.uploads_nofiles')));

        }

    },

    /**
    * The uploader queue may fire an "alluploadscomplete" event
    * even though there still are aborted/otherwise unfinished uploads
    *
    * Do a check for unfinished uploads in the UI
    *
    * @method _dispatchComplete
    * @protected
    */
    _dispatchComplete: function () {

        if(this.getFilesNotUploaded().length > 0) {

            this._partialComplete();

        } else {

            this._complete();

        }

    },

    _totalUploadProgress: function (event) {

        // if we have uploaded bytes and totalbytes, duck punch event
        // https://github.com/yui/yui3/issues/1176
        if(event.bytesLoaded && event.bytesTotal) {
            event.percentLoaded = Math.min(100, Math.round(10000*event.bytesLoaded/event.bytesTotal) / 100);
        }

        var content,
            totals,
            p = event.percentLoaded;

        if(p) {

            // for numerical values
            if(!isNaN(parseFloat(p)) && isFinite(p)) {

                content = Y.Lang.sub(this.get('strings.total_uploaded'), { uploaded: p });

            } else {

                content = this.get('strings.uploading');

            }

        } else if (this.getHavePendingFiles()) {

            totals = this.getFilesTotals();

            content = Y.Lang.sub(this.get('strings.uploads_pending'), { count: totals.count, size: this.getFriendlySize(totals.size) });

        } else if (this.getHaveNoFiles()) {

            content = this.get('strings.uploads_empty');
        }

        if(content) {

            this.set('status', content);

        }

    },


    /**
    * All files complete handler
    *
    * @method _complete
    * @protected
    */
    _complete: function () {

        this.set('status', this.get('strings.uploads_complete'));
        this.fire('fullycomplete');

    },

    /**
    * Stalled files handler
    *
    * @method _uploadStalled
    * @protected
    */
    _uploadStalled: function () {

        this.set('status', Y.Lang.sub(this.get('strings.stalled')));

    },

    /**
    * Some files complete handler
    *
    * @method _partialComplete
    * @protected
    */
    _partialComplete: function () {

        this.set('status', this.get('strings.partial_complete'));
        this.fire('partialcomplete');
    },

    /**
    * Starts uploading all files
    *
    * @method start
    * @public
    */
    start: function () {
        this.uploadAll();

        // duck punch broken uploader (uploader-queue)
        this.queue.on('uploadcacel', function () {
            this.fire('uploadcancel');
        }, this);
    },

    /**
    * Cancels queue and fires the incomplete event
    *
    * @method cancelAllUploads
    * @public
    */
    cancelAllUploads: function () {

        if (this.queue) {
            this.queue.cancelUpload();
        }
        this.fire('incomplete');

    },

    cancelFile: function (fileToCancel) {

        // set fileList to array without file to be deleted
        this.set('fileList', Y.Array.filter(this.get('fileList'), function (file) {
            return file.get('id') !== fileToCancel.get('id');
        }));

        // if a queue is active
        if (this.queue) {

            // cancel upload, remove from queue
            this.queue.cancelUpload(fileToCancel);

            // the next two lines should be handled by YUI Uploader Queue, but is not
            // monkeypatching in the mean time
            // reduce totalBytes
            this.queue.totalBytes -= fileToCancel.get('size');

            // remove the file's uploaded bytes from uploaded bytes total
            this.queue.totalBytesUploaded = Math.max(0, this.queue.totalBytesUploaded - fileToCancel.get('bytesUploaded'));

            // if we cancelled the last item in the queue
            // TODO: use existing methods
            if(Y.Object.isEmpty(this.queue.currentFiles) && this.queue.queuedFiles.length === 0 && this.getFilesNotUploaded().length === 0) {
                this.cancelAllUploads();
            }
        }

        this.fire('filecancel', {
            file: fileToCancel
        });
    },

    // main cancel path, used in X-buttons and Cancel-buttons
    cancel: function () {

        this._confirmCancel();

    },

    retryFile: function (file) {

        var q = this.queue;

        // if no queue at all, use uploadThese
        if(!q) {
            this.uploadThese([file], this.get('url'));
        } else {
            // if the file was cancelled by error, re-register
            if(!q.currentFiles.hasOwnProperty(file.get('id'))) {
                q._registerUpload(file);
            }

            q.forceReupload(file);
        }

    },

    getFile: function (id) {

        return Y.Array.find(this.get('fileList'), function (file) {
            return file.get('id') === id;
        });

    },

    getFileError: function (event) {

       var source = event.source === 'io' ? this.get('strings.error_network') : this.get('strings.error_http'),
           msg,
           test500 = /5\d+/,
           test400 = /4\d+/;

        switch(event.status) {

            case test500.test(event.status):
                msg = Y.Lang.sub(this.get('strings.error_server'), {
                    status: event.status
                });
                break;
            case test400.test(event.status):
                msg = Y.Lang.sub(this.get('strings.error_server'), {
                    status:         event.status,
                    status_text:    event.statusText
                });
                break;
            case 0:
                msg = this.get('strings.error_aborted');
                break;
            default:
                msg = Y.Lang.sub(this.get('strings.error_generic'), {
                    status:         event.status,
                    status_text:    event.statusText,
                    source:         source
                });
                break;
        }

        return {
            source: source,
            message: msg
        };

    },

    getUploadsInProgress: function () {

        return Y.Array.filter(this.get('fileList'), function (file) {
            return file.get('status') === UPLOADING || file.get('status') === STARTED;
        });

    },

    getHaveUploadsInProgress: function () {

        return this.getUploadsInProgress().length > 0;

    },


    getPendingFiles: function () {

        return Y.Array.filter(this.get('fileList'), function (file) {
            return typeof file.get('status') === 'undefined';
        });

    },

    getHavePendingFiles: function () {

        return this.getPendingFiles().length > 0;

    },

    getHaveNoFiles: function () {

        return !this.queue && this.get('fileList').length === 0;

    },

    getFilesTotals: function () {

        var filesCount      = 0,
            filesTotalSize  = 0;

        Y.each(this.get('fileList'), function (file) {
            filesCount++;
            filesTotalSize += file.get('size');
        });

        return {
            count: filesCount,
            size: filesTotalSize
        };

    },

    getFilesNotUploaded: function () {

        return Y.Array.filter(this.get('fileList'), function (file) {
            return file.get('status') !== COMPLETE;
        });

    },

    /**
    * Convert number of bytes into human readable string (ala "2 MB")
    * @method getFriendlySize
    * @param size the size in bytes
    *
    * https://github.com/sspencer/yui3-gallery
    * @param size {Number}
    * @method getFriendlySize
    * @protected
    */
    getFriendlySize: function(size) {

        var i, units = [
            this.get("strings.size_b"),
            this.get("strings.size_kb"),
            this.get("strings.size_mb"),
            this.get("strings.size_gb"),
            this.get("strings.size_tb")
        ];

        if(!size) {
            return 0;
        }

        for (i = 0; size > 1024; i++) { size /= 1024; }

        if (i < units.length) {
            return Math.round(size*10)/10 + units[i];
        } else {
            return "?";
        }
    }

}, {

    ATTRS: {


        dragAndDropArea: { value: '.yui3-dd-proxy' },

        errorAction: { value: Y.Uploader.Queue.RESTART_ASAP },

        strings: {

            value: {


                cancel:             'Cancel',
                cancel_pending:     'Discard {count} pending files?',
                cancel_remaining:   'Stop uploading and discard {count} remaining files?',

                done:               'Done',
                drop_pending:       'Files detected, drop them here!',
                drop_ready:         'Drag and drop files here',

                error:              '{error} error',
                error_http:         'Server',
                error_network:      'Network',

                error_aborted:      'Upload request aborted.',
                error_client:       'Upload failed due to a client error {status}: {status_text}.',
                error_generic:      'Upload failed due to an {source} error. Details: {status_text} code {status}.',
                error_server:       'Upload failed due to a server error {status}.',

                not_started:        'Not started',

                partial_complete:   'Some files were not uploaded - <a class="yui3-dismiss">dismiss</a>',

                size_b:             ' B',
                size_kb:            ' KiB',
                size_mb:            ' MiB',
                size_gb:            ' GiB',
                size_tb:            ' TiB',
                stalled:            '<p class="yui3-stalled">Ouch, looks like the current uploads have stalled - <a class="yui3-skip">continue</a></p>',

                time_s:             'second',
                time_m:             'minute',
                time_h:             'hour',
                time_d:             'day',
                time_u:             'unknown time',
                time_plural:        's',
                time_remaining:     '<span>{time} remaining</span>',
                total_uploaded:     'Total uploaded: <strong class="yui3-percentage">{uploaded}%</strong>',

                upload:             'Upload',
                upload_complete:    'Upload complete',
                uploading:          'Uploading',
                uploads_complete:   'Uploads complete!',
                uploads_empty:      'No files selected for upload.',
                uploads_nofiles:    'There were no files to upload. Cancel and try again?',
                uploads_pending:    'Will upload <strong>{count}</strong> files, <strong>{size}</strong> in total.'

            }
        },

        /**
        * Uploader status presented as a string
        *
        * @attribute status
        * @type {String}
        * @default
        */
        status: { value: '' },

        /**
        * Files drop area status as a string
        *
        * @attribute dropAreaStatus
        * @type {String}
        * @default
        */
        dropAreaStatus: { value: '' },

        /**
        * Time remaining as a string
        *
        * @attribute timeRemaining
        * @type {String}
        * @default
        */
        timeRemaining: { value: '' }

    }


});

Y.namespace('U1.Widgets').Uploader = Uploader;
